int Search::calcExtensions(const Board &board,
                           NodeInfo *node, NodeInfo *parentNode,
                           CheckStatusType in_check_after_move,
                           int moveIndex,
                           Move move) {
   // see if we should apply any extensions at this node.
   int depth = node->depth;
   node->extensions = 0;
   int extend = 0;
   int pruneOk = board.checkStatus() != InCheck;
   score_t swap = Scoring::INVALID_SCORE;
   bool badCheck = false;
   if (in_check_after_move == InCheck) { // move is a checking move
      // extend if check does not lose material or is a discovered check
      if ((swap = seeSign(board,move,0)) ||
          board.isPinned(board.oppositeSide(),move)) {
          node->extensions |= CHECK;
#ifdef SEARCH_STATS
          controller->stats->check_extensions++;
#endif
          extend += node->PV() ? PV_CHECK_EXTENSION : NONPV_CHECK_EXTENSION;
      }
      else {
         // Bad checks can be reduced or pruned but only at low depth
         badCheck = true;
      }
   }
   if (passedPawnPush(board,move)) {
      node->extensions |= PAWN_PUSH;
      extend += PAWN_PUSH_EXTENSION;
#ifdef SEARCH_STATS
      controller->stats->pawn_extensions++;
#endif
   }
   else if (TypeOfMove(move) == Normal &&
            Capture(move) != Empty && Capture(move) != Pawn &&
            board.getMaterial(board.oppositeSide()).pieceCount() == 1 &&
            board.getMaterial(board.sideToMove()).noPieces()) {
      // Capture of last piece in endgame.
      node->extensions |= CAPTURE;
      extend += CAPTURE_EXTENSION;
#ifdef SEARCH_STATS
      ++controller->stats->capture_extensions;
#endif
   }
   if (extend) {
      return std::min<int>(extend,DEPTH_INCREMENT);
   }

   // See if we do late move reduction. Moves in the history phase of move
   // generation can be searched with reduced depth.
   if (depth >= LMR_DEPTH && moveIndex >= 1+2*node->PV() &&
       ((node->ply == 0 && !CaptureOrPromotion(move))  || GetPhase(move) == MoveGenerator::HISTORY_PHASE || (GetPhase(move) != MoveGenerator::WINNING_CAPTURE_PHASE && board.checkStatus() == InCheck && moveIndex >=3)) &&
       !passedPawnMove(board,move,6)) {
      extend -= LMR_REDUCTION[node->PV()][depth/DEPTH_INCREMENT][std::min<int>(63,moveIndex)];
      if (extend) {
         node->extensions |= LMR;
#ifdef SEARCH_STATS
         ++controller->stats->reduced;
#endif
      }
   }

   // for pruning decisions, use depth after LMR
   const int predictedDepth = depth + extend;

   pruneOk &= !node->PV() && parentNode->num_try &&
       Capture(move) == Empty &&
       TypeOfMove(move) == Normal &&
       (!badCheck || predictedDepth <= 3*DEPTH_INCREMENT) &&
       !passedPawnMove(board,move,5) &&
       !Scoring::mateScore(parentNode->alpha);

   if (pruneOk) {
      // do not use predictedDepth for LMP
      if(depth/DEPTH_INCREMENT <= LMP_DEPTH &&
         in_check_after_move == NotInCheck &&
         GetPhase(move) >= MoveGenerator::HISTORY_PHASE &&
         moveIndex >= LMP_MOVE_COUNT[depth/DEPTH_INCREMENT]) {
#ifdef SEARCH_STATS
         ++controller->stats->lmp;
#endif
#ifdef _TRACE
         indent(node->ply); cout << "LMP: pruned" << endl;
#endif
         return PRUNE;
      }
      // futility pruning, enabled at low depths
      if (predictedDepth <= FUTILITY_DEPTH) {
         // Threshold was formerly increased with the move index
         // but this tests worse now.
         score_t threshold = parentNode->beta - FUTILITY_MARGIN[predictedDepth/DEPTH_INCREMENT];
         if (node->eval == Scoring::INVALID_SCORE) {
            node->eval = node->staticEval = scoring.evalu8(board);
         }
         if (node->eval < threshold) {
#ifdef SEARCH_STATS
            controller->stats->futility_pruning++;
#endif
#ifdef _TRACE
            indent(node->ply); cout << "futility: pruned" << endl;
#endif
            return PRUNE;
         }
      }
   }
   // See pruning. Losing captures and moves that put pieces en prise
   // are pruned at low depths. Losing checks can be pruned.
   if (predictedDepth <= SEE_PRUNING_DEPTH &&
       !node->PV() && parentNode->num_try &&
       !Scoring::mateScore(node->alpha) &&
       !badCheck &&
       GetPhase(move) > MoveGenerator::WINNING_CAPTURE_PHASE) {
       if (swap == Scoring::INVALID_SCORE) swap = seeSign(board,move,0);
       if (!swap) {
#ifdef SEARCH_STATS
          ++controller->stats->see_pruning;
#endif
          return PRUNE;
       }
   }
   return extend;
}

